/*
 * Copyright (C) 2017 Mark Hills <mark@pogo.org.uk>
 *
 * Listen for an asynchronous event from a PostgreSQL server, and call
 * the given command when it happens.
 */

#define _GNU_SOURCE /* asprintf() */
#include <assert.h>
#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <syslog.h>
#include <string.h> /* strsignal() */
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "libpq-fe.h"

#define MINUTE 60

void lerror(const char *prefix)
{
	syslog(LOG_ERR, "%s: %s", prefix, strerror(errno));
}

int fork_exec(const char *path, char *const argv[])
{
	int status;
	pid_t pid;

	pid = fork();

	if (pid == -1) {
		lerror("fork");
		return -1;
	}

	if (pid == 0) { /* child */
		execvp(path, argv);
		lerror("execvp");
		exit(EX_OSERR);
	}

	if (waitpid(pid, &status, 0) == -1) {
		lerror("waitpid");
		return -1;
	}

	if (WIFEXITED(status) && WEXITSTATUS(status) == 0)
		return 0;

	if (WIFSIGNALED(status)) {
		syslog(LOG_ERR, "Child process terminated by signal %s",
			strsignal(WTERMSIG(status)));
	}

	if (WIFEXITED(status) && WEXITSTATUS(status) != 0) {
		syslog(LOG_ERR, "Child process exited with return code %d",
			WEXITSTATUS(status));
	}

	syslog(LOG_ERR, "Child process exited abnormally");
	return -1;
}

/*
 * Connect to the database and start listening
 *
 * Return: NULL if unable to connect and listen, otherwise a
 * valid PostgreSQL connection with listener
 */

static PGconn *listen(const char *cs, const char *event)
{
	char *sql;
	PGconn *conn;
	PGresult *res;

	/* Make the connection */

	conn = PQconnectdb(cs);
	if (PQstatus(conn) != CONNECTION_OK) {
		syslog(LOG_ERR, "%s", PQerrorMessage(conn));
		return NULL;
	}

	/* Send the initial LISTEN command */

	(void)asprintf(&sql, "LISTEN %s", event);
	res = PQexec(conn, sql);
	if (PQresultStatus(res) != PGRES_COMMAND_OK) {
		syslog(LOG_ERR, "%s", PQerrorMessage(conn));
		conn = NULL;
	}

	PQclear(res);
	free(sql);

	return conn;
}

/*
 * Wait for a notification and launch the provided command
 *
 * Return: -1 of we are not connected, otherwise the number of events
 * which may be zero
 */

static unsigned int hold(PGconn *conn)
{
	int sd;
	unsigned int nevents = 0;
	fd_set input_mask;
	PGnotify *notify;

	assert(conn);

	/* Prepare the data needed for select() */

	sd = PQsocket(conn);
	if (sd == -1) {
		syslog(LOG_ERR, "%s", PQerrorMessage(conn));
		return -1;
	}

	FD_ZERO(&input_mask);
	FD_SET(sd, &input_mask);

	if (select(sd + 1, &input_mask, NULL, NULL, NULL) == -1) {
		lerror("select");
		return -1;
	}

	if (PQconsumeInput(conn) == 0) {
		syslog(LOG_ERR, "%s", PQerrorMessage(conn));
		return -1;
	}

	while ((notify = PQnotifies(conn)) != NULL) {
		PQfreemem(notify);
		nevents++;
	}

	return nevents;
}

/*
 * Select a random sleep value, honouring the "lower" limit
 */

static unsigned int between(unsigned int lower, unsigned int desired)
{
	if (desired <= lower)
		return desired;

	return lower + rand() % (desired - lower);
}

/*
 * Apply a rate limit on faults
 */

static void ratelimit(void)
{
	const unsigned int grace = 1 * MINUTE, truncate = 1 * MINUTE;
	static unsigned int x = 0;
	static struct timespec last, now;

	/*
	 * If nothing happened for a while, assume everything is ok
	 */

	if (clock_gettime(CLOCK_MONOTONIC, &now) == -1)
		abort();

	if (x > 0 && now.tv_sec > last.tv_sec + grace) {
		syslog(LOG_INFO, "Resetting timeout");
		x = 0;
	}

	/*
	 * Sleep
	 */

	if (x > 0) {
		unsigned int a;

		a = between(1, x);

		syslog(LOG_INFO, "Sleeping %u %s before retry",
			a, a == 1 ? "second" : "seconds");

		sleep(a);
	}

	/*
	 * Truncated exponential backoff
	 */

	if (x == 0) {
		x = 1;
	} else {
		x *= 2;
	}

	if (x > truncate)
		x = truncate;

	last = now;
}

/*
 * Listen to database with connection string "cs" for the given
 * "event", and run the given command
 */

static void forever(const char *cs, const char *event,
		const char *path, char *const argv[])
{
	for (;;) {
		PGconn *conn;

		conn = listen(cs, event);
		if (!conn)
			goto again;

		/*
		 * Every time we connect, we know we may have lost
		 * messages, so we run the command to ensure we are
		 * up-to-date.
		 */

		if (fork_exec(path, argv) != 0)
			ratelimit();

		for (;;) {
			unsigned int events;

			events = hold(conn);
			if (events == -1)
				break;

			if (events == 0)
				continue;

			if (fork_exec(path, argv) != 0)
				ratelimit();
		}

		/*
		 * Any socket related failure and we force
		 * a start again from clean
		 */

		PQfinish(conn);
		conn = NULL;

	again:
		ratelimit();
	}
}

void usage(FILE *fd)
{
	fprintf(fd, "Usage: pgnotifyd [-b] [-d <conninfo>] <event> [--] <command> [ <arg> ... ]\n");
}

int main(int argc, char *argv[])
{
	int opt;
	const char *event;
	char *cs = "";
	bool bg = false;

	for (;;) {
		opt = getopt(argc, argv, "bd:h");
		if (opt == -1)
			break;

		switch (opt) {
		case 'd':
			cs = optarg;
			break;
		case 'b':
			bg = true;
			break;
		case 'h':
			usage(stdout);
			return EX_OK;
		default:
			usage(stderr);
			return EX_USAGE;
		}
	}

	argv += optind;
	argc -= optind;

	if (argc < 1) {
		usage(stderr);
		return EX_USAGE;
	}

	event = argv[0];

	/* Syslog */

	openlog("pgnotifyd", LOG_PID | LOG_PERROR, LOG_DAEMON);

	/* Go to the background if required */

	if (bg) {
		if (daemon(0, 0) == -1) {
			lerror("daemon");
			return EX_OSERR;
		}
	}

	forever(cs, event, argv[1], &argv[1]);

	closelog();

	return EX_OK;
}
